proc _deet_apply {action file line char} {
	global gdbInfo gdbBreakpoint

	switch $action {
		list	{return [list $file $line $char]}
		set {
			if {[sh_vnull gdbBreakpoint($file:$line)]} {
				if {$gdbInfo(func) == ""} {
					sendCommand "b $file:$line"
				} else {
					sendCommand "b $gdbInfo(func)"
				}
				if {$gdbInfo(bpnum) != "" } {
				  set gdbBreakpoint($file:$line) $gdbInfo(bpnum)
				}
			}
		}
		delete {
			sendCommand "delete $gdbBreakpoint($file:$line)"
			set gdbBreakpoint($file:$line) ""
		}
	}
}

proc deet_breakpoint {action file line char} {
	global gdbInfo gdbBreakpoint
	set result ""

	if {$gdbInfo(running)} {error "Program running"}
	switch -glob -- $action {
		{-s*} {
			set action set
			if {$file == "" && $line == "0" && $char == "0"} {
				sendCommand "b main"
				set gdbInfo(contcmd) "step"
				return
			}
		}
		{-d*} {
			set action delete
			if {$file == "" && $line == "0" && $char == "0"} {
				sendCommand "delete"
				foreach i [array names gdbBreakpoint] {
					unset gdbBreakpoint($i)
				}
				set gdbInfo(contcmd) "cont"
				return
			}
		}
		{-l*} {set action list}
	}

	set gdbInfo(contcmd) "cont"

	if {$file != ""} {
		set files $file
	} {
		set files $gdbInfo(files)
	}
	foreach i $files {
		if { $line != 0 } {
			sendCommand "info line $i:$line"
			if {$gdbInfo(error) == "" } {
				lappend result [_deet_apply $action $i $line 1]
			}
			if {$gdbInfo(error) == "NO_LINE" && $file != ""} {
				error
				return
			}
			continue
		}
		for {set num 1} 1 {incr num} {
			sendCommand "info line $i:$num"
			if {$gdbInfo(error) == "NO_CODE"} continue
			if {$gdbInfo(error) == "NO_LINE"} break
			append result [_deet_apply $action $i $num 1]
			incr num
		}
	}
	gdbDebugPrint "DD: $result"
	return $result
}

proc deet_continue {{opt -c} {cmd ""}} {
	global gdbInfo gdbFile
	set gdbInfo(breakpoint) ""
	switch -- $opt {
	   "-i"  {
		if {! $gdbInfo(running)} return
		set gdbInfo(running) 0
		if {$gdbInfo(childpid) != ""} {
			puts "killing $gdbInfo(childpid)"
			exec kill -INT $gdbInfo(childpid)
			waitForPrompt
		}
		return
	   }
	   "-c" {set gdbInfo(contcmd) cont}
	   "-s" {set gdbInfo(contcmd) step}
	   "-n" {set gdbInfo(contcmd) next}
	}
	if {$gdbInfo(running)} {error "Program already running"}
	set gdbInfo(running) 1
	sendCommand $gdbInfo(contcmd) 0
	vwait gdbInfo(running)
}

proc deet_frame {{n ""}} {
	global gdbInfo 
	set gdbInfo(frame) "-1"
	set gdbInfo(error) ""

	if {$gdbInfo(running)} {error "Program running"}

	if {$n == ""} {
		sendCommand "up 0"
		if {$gdbInfo(error) == "NO_STACK"} {
			error "no current frame"
		}
		set file [shpat_%  $gdbInfo(breakpoint) :.*]
		set line [shpat_## $gdbInfo(breakpoint) .*:]
		return [list $gdbInfo(frame) $gdbInfo(func) $file $line 1]
	} else {
		sendCommand "frame $n"
		if {$gdbInfo(frame) != $n} {
			error "no such frame: $n"
		}
	}
}
