source $deetnub_library/gdbSym.tcl
source $deetnub_library/gdbCntl.tcl

global gdbInfo gdbBreakpoint gdbFile

proc gdbInit {} {
	global gdbInfo
	set gdbInfo(debug)	0	;# flag to debug gdb messages
	set gdbInfo(error)	""	;# error from gdb command
	set gdbInfo(func)	""	;# function found in gdb command
	set gdbInfo(line)	""	;# line number found in gdb command
	set gdbInfo(file)	""	;# file name found in gdb command
	set gdbInfo(address)	""	;# address found in gdb command
	set gdbInfo(string)	""	;# string for printed char * 
	set gdbInfo(frame)	""	;# frame number found in gdb command
	set gdbInfo(bpnum)	""	;# breakpoint number found in command
	set gdbInfo(breakpoint)	""	;# breakpoint info found in command
	set gdbInfo(members)	""	;# members of c structure
	set gdbInfo(type)	""	;# type name found in gdb command
	set gdbInfo(childpid)	""	;# process id of running program
	set gdbInfo(target)	""	;# Name of executable
	set gdbInfo(args)	""	;# Arguments that gdb is invoked with
	set gdbInfo(vars)	""	;# list of variables for command
	set gdbInfo(files)	""	;# list of files in program
	set gdbInfo(contcmd)	cont	;# Name of command sent to continue
	set gdbInfo(running)	0	;# If program is running
}

proc gdbDebugPrint {cmd} {
	global gdbInfo
	if {! $gdbInfo(debug)} return
	if {! [winfo exists .gdbout]} {
		toplevel .gdbout
		text_create .gdbout.t
		pack .gdbout.t
		wm iconify .gdbout
	}
#	puts stderr "[lindex [info level -1] 0]: $cmd"
	text_add .gdbout.t "[lindex [info level -1] 0]: $cmd\n"
	.gdbout.t.t see end
}

proc waitForPrompt {} {
	global gdbFile gdbInfo
	while {! [eof $gdbFile]} {
		set reply [gets $gdbFile]
		gdbDebugPrint $reply
		switch -regexp $reply {
			.*__GDB_PROMPT__	{return}
			{Source files .*}	{handleSources ; break}
                        {Line [0-9]+ .*}        {handleLine "$reply"}
                        {Breakpoint.* at .*:.*} {handleBp "$reply"}
                        {type = .*}              {handleType "$reply"}
                        {#[0-9]* .*}		{handleFunc "$reply"}
                        {\$.* = .*}		{handleAddress "$reply"}
                        {.* = .*}		{handleVar "$reply"}
                        {All .* matching.*:}	{handleFind "$reply" ; return}
                        {All defined.*}		{handleFind "$reply" ; return}
                        {.*Using the running.*child.*}
						{handlePid "$reply"}
                        {(No line|Line number.* out of range).*}
						{set gdbInfo(error) "NO_LINE"}
                        {Cannot access memory.*} {set gdbInfo(error) "NO_MEM"}
			{.*File format not recognized.*}
						{set gdbInfo(error) "NO_PROG"}
                        {The program is not being run.*}
                                                {set gdbInfo(error) "NOT_RUN"}
                        {.*No such file.*}	{set gdbInfo(error) "NO_FILE"}
                        {No source file.*}	{set gdbInfo(error) "NO_FILE"}
                        {.*contains no code.*}	{set gdbInfo(error) "NO_CODE"}
                        {.*No symbol.*}		{set gdbInfo(error) "NO_SYM"}
                        {No stack\..*}		{set gdbInfo(error) "NO_STACK"}
                        {^No frame.*}		{set gdbInfo(error) "NO_FRAME"}
                        {Program exited.*}	{set gdbInfo(error) "NO_PROG"
						 set gdbInfo(running) -1}
                        {.*\(y or n\)$}		{sendCommand "y"; return}
		}
		if {$gdbInfo(error) != ""} {gdbDebugPrint "Err:$gdbInfo(error)"}
	}
}

proc sendCommand {cmd {wait 1}} {
        global gdbInfo gdbFile
	set gdbInfo(error) ""
	set gdbInfo(bpnum) ""
	gdbDebugPrint $cmd
	puts $gdbFile $cmd
	flush $gdbFile
	if {$wait} {waitForPrompt}
}

##########

proc handleSources {} {
	global gdbFile gdbInfo
	set on 0
	set gdbInfo(files) ""
	while {! [eof $gdbFile]} {
		set result [gets $gdbFile]
		gdbDebugPrint $result
		switch -glob $result {
			*__GDB_PROMPT__		{return}
			{Source files *}	{}
			default {
				foreach i $result {
					set f [string trimright $i ,]
					gdbDebugPrint "FILE: $i"
					if [string match *.c $f] {
						lappend gdbInfo(files) $f
					}
				}
			}
		}
	}
}

proc handleLine {line} {
	global gdbInfo
	set result $line
	set gdbInfo(func) ""
	switch -glob $result {
		{*contains no code*}	{set gdbInfo(error) NO_CODE}
		default	{
			set gdbInfo(line) [lindex $result 1]
			set gdbInfo(file) [lindex $result 3]
			set name [lindex $result 8]
			gdbDebugPrint "$gdbInfo(file) $gdbInfo(line)"
			if [regexp {<[^+]+>} $name] {
				set name [string trimleft $name <]
				set gdbInfo(func) [string trimright $name >]
			}
		}
	}
}
		
proc handleBp {bp} {
	global gdbInfo
	set gdbInfo(bpnum) [string trimright [lindex $bp 1] ,]
	if [string match [lindex $bp 2] "at"] {
		set loc "[string trimright [lindex $bp 5] ,]:[string trimright [lindex $bp 7] .]"
	} {
		set loc [shpat_## $bp {.* }]
	}
	set gdbInfo(breakpoint) $loc
}

proc handleVar {v} {
	global gdbInfo
	gdbDebugPrint [lindex $v 0]
	lappend gdbInfo(vars) [lindex $v 0]
}

proc handleAddress {a} {
	global gdbInfo
	set gdbInfo(string) ""
	if {[string match {*"*} $a]} {
		if {[regexp {([^"]*)(".*)} $a dummy begin end]} {
			set a [string trim $begin]
			set gdbInfo(string) $end
		}
	}
	set gdbInfo(address) [shpat_## $a {.* }]
	gdbDebugPrint "$gdbInfo(address), $gdbInfo(string)"
}

proc handleType {t} {
	global gdbInfo gdbFile

	regsub {^type = } $t "" gdbInfo(type)
	regsub " " $gdbInfo(type) = gdbInfo(type)
	set gdbInfo(members) ""
	
	if {[string match "*\{" $t]} {
		while {! [eof $gdbFile]} {
			set result [string trim [gets $gdbFile]]
			gdbDebugPrint $result
			if {[string match "\}*" $result]} break
		        if {[string match "*();" $result]} {
		                regsub {(.*\(\*)([a-zA-Z0-9_]+)(.*);} \
					$result {\1\3} tp
				regsub {(.*\(\*)([a-zA-Z0-9_]+)(.*);} \
					$result {\2} nm
		        } else {
				regsub {(.* \**)([a-zA-Z0-9_]+)(.*);} \
					 $result {\1\3} tp
		                regsub {(.* \**)([a-zA-Z0-9_]+)(.*);} \
					$result {\2} nm
			}       
			regsub -all " " [string trim $tp] "=" tp
			lappend gdbInfo(members) [list $tp $nm]
		}
	}
}

proc handleFunc {l} {
	global gdbInfo

	set gdbInfo(breakpoint) ""
	if {[regexp {.*\(.*\) at .*} $l]} {
		set gdbInfo(breakpoint) [shpat_## $l {.*at }]
	}
	set gdbInfo(frame) [shpat_# [lindex $l 0] {#}]
	gdbDebugPrint $gdbInfo(frame)
	if {[string match 0x* [lindex $l 1]]} {
		set gdbInfo(func) [lindex $l 3]
	} else {
		set gdbInfo(func) [lindex $l 1]
	}
}

proc handleFind {arg} {
	global gdbFile gdbInfo

	while {! [eof $gdbFile]} {
		set result [gets $gdbFile]
		gdbDebugPrint $result
		switch -glob $result {
			*__GDB_PROMPT__		{return}
			{File *} {
				set gdbInfo(file) [shpat_% [lindex $result 1] :]
			}
			{*;} {
				set name [shpat_## $result {.* \*+}]
				set name [shpat_%% $name {(\(\))+;}]
				lappend gdbInfo(func) "$gdbInfo(file):$name"
			}
		}
	}
}

proc handlePid {reply} {
	global gdbInfo

	set pid [shpat_## $reply {.* }]
	set gdbInfo(childpid) [shpat_% $pid {\.}]
	gdbDebugPrint $gdbInfo(childpid)
}

proc handleReady {} {
	global gdbInfo
	if {$gdbInfo(running)} {
		waitForPrompt
		if {$gdbInfo(error) == "NOT_RUN"} {
		   if {$gdbInfo(contcmd) != "cont"} {
			sendCommand "b main"
		   }
        	   sendCommand "set args $gdbInfo(args)"
        	   sendCommand run 0
		} else {
		   if {$gdbInfo(childpid) == ""} {sendCommand "info target"}
		   set gdbInfo(running) 0
		}
	}
}

if {[info commands vwait] == ""} {
	proc vwait {var} {tkwait variable $var}
}

proc spawngdb {} {
	global gdbFile gdbInfo deetnub_library
	gdbDebugPrint "Starting gdb"
	if {[file exists $deetnub_library/sess]} {
		set gdbFile [open "|$deetnub_library/sess gdb -nx -q -tty=[exec tty] 2>@stdout" w+]
	} else {
		set gdbFile [open "|gdb -nx -q -tty=[exec tty] 2>@stdout" w+]
	}
	catch {fconfigure $gdbFile -buffering none}
	catch {fileevent $gdbFile readable handleReady}
	sendCommand "set prompt __GDB_PROMPT__\\n"
	sendCommand "set confirm off"
	sendCommand "set width 0"
}

proc deet_close {} {
	global gdbFile
	if {[info exists gdbInfo(running)]} {
		deet_continue -i
		close $gdbFile
	}
}

proc deet_open {target dargs} {
	global gdbInfo
	if {! [info exists gdbInfo(running)]} {
		gdbInit
		spawngdb
	}
	if {$gdbInfo(target) != $target} {
		sendCommand "delete"
		sendCommand "file $target"
		if {($gdbInfo(error) == "NO_PROG") ||
		    ($gdbInfo(error) == "NO_FILE")} {
			error "Error loading $target"
		}
	} else {
		sendCommand "kill"
	}
	set gdbInfo(target) $target
	set gdbInfo(args) $dargs
}
