proc setframe {n {w ""}} {
	global DeetFrame
	set oldFrame $DeetFrame

	deet_frame $n
	set f [deet_frame]
	# Be sure to catch this proc because the above could fail
	set DeetFrame $n
	if {$w != ""} {
		debugwin_setfile $w [lindex $f 2] [lindex $f 3]
	}
	if {[winfo exists .stack]} {
		set w .stack.h
		set h [hlist_subwidget $w text]
		$h tag configure $oldFrame -foreground "" -background ""
		$h tag configure $n -foreground red -background white
	}
}

proc up {{l 1} {w ""}} {
	global DeetFrame
	if {$DeetFrame != ""} {
		catch {setframe [expr [lindex $DeetFrame 0] + $l] $w}
	}
}

proc down {{l 1} {w ""}} {
	global DeetFrame
	if {$DeetFrame != ""} {
		catch {setframe [expr [lindex $DeetFrame 0] - $l] $w}
	}
}

proc xwhere {win} {
	set w .stack
	if {[winfo exists $w]} return

	toplevel $w
	window_title $w "Stack"
	hlist_create $w.h
        buttons_create $w.b Close:r Down:r Up:r Goto:r
        buttons_command $w.b 1 "hlist_clear $w.h ; destroy $w"
        buttons_command $w.b 2 "down 1 $win"
        buttons_command $w.b 3 "up 1 $win"
        buttons_command $w.b 4 "xwhere_goto $w.h $win"
	mpack $w.b:bx $w.h:ez
	xwhere_stack 
}

proc xwhere_stack {} {
	global DeetFrame

	set w .stack
	if {! [winfo exists $w]} return
	hlist_clear $w.h
	for {set i 0} {1} {incr i} {
		if {[catch {deet_frame $i}]} break
		set f [deet_frame]
		set num [lindex $f 0]
		set name [lindex $f 1]
		set file [lindex $f 2]
		set line [lindex $f 3]
		set char [lindex $f 4]
		if {$file != ""} {
			append name " (line $line.$char of $file)"
		}
		hlist_insert $w.h $i "$num: $name"
		hlist_action $w.h $i "xwhere_expand $w.h $i"
		hlist_seticon $w.h $i plus
		if {$num == $DeetFrame} {
   			[hlist_subwidget $w.h text] tag configure \
				$i -foreground red -background white
		}
	}
	[hlist_subwidget $w.h text] see 0.0
	catch {deet_frame $DeetFrame}
}

proc xwhere_expand {w num} {
	set i 0
	if {[catch {deet_frame $num}]} return
	set p [concat [deet_sym -params] [deet_sym -locals]]
	while {$p != ""} {
 	  set n [shpat_## [lindex $p 0] .*:]
	  vlist_add $w $num/$i $n [lindex $p 1] [lindex $p 2]
	  set p [lrange $p 3 end]
	  incr i
	}
	if {$i == "0"} {
		hlist_seticon $w $num none
	}

}


proc xwhere_goto {w win} {
	global DeetFrame
	set n [hlist_selection $w]
	if {[string match */* $n ]} return
	catch {setframe $n $win}
}

proc where {} {
	global DeetFrame DeetStack DeetLastStack DeetSeqNum
	set result ""
	
	if {$DeetFrame == ""} {
		error "Program not yet run"
	}

	if { $DeetLastStack != $DeetSeqNum} getStack

	for {set i 0} {$DeetStack($i) != ""} {incr i} {
		if {$i == $DeetFrame} {
			append result "* $DeetStack($i)\n"
		} else {
			append result "  $DeetStack($i)\n"
		}
	}
	return $result
}

proc getStack {} {
	global DeetStack DeetLastStack DeetSeqNum DeetFrame
	set i 0

	while 1 {
		if {[catch {deet_frame $i}]} break
		set f [deet_frame]
		set num [lindex $f 0]
		set name [lindex $f 1]
		set file [lindex $f 2]
		set line [lindex $f 3]
		set char [lindex $f 4]
		set result "$num ${name}("
		set p [deet_sym -params]
		while {$p != ""} {
	 	   set n [shpat_## [lindex $p 0] .*:]
		   append result "$n=[deet_getval [lindex $p 1] [lindex $p 2]], "
		   set p [lrange $p 3 end]
		}
		set DeetStack($i) "[shpat_% $result ", "]) at $file:$line.$char"
		incr i
	}
	set DeetStack($i) ""
	set DeetLastStack $DeetSeqNum
	deet_frame $DeetFrame
}
