##############################
# sym.tcl: Symbol management #
##############################

proc vlist_setup {w} {
	hlist_create $w.h
	entry_create $w.e "Expression:" "Display"
	buttons_create $w.b Close:r Watch:r Modify:r Remove:r
        buttons_command $w.b 1 "hlist_clear $w.h ; destroy $w"
        buttons_command $w.b 2 "vlist_watch $w \[hlist_selection $w.h\]"
        buttons_command $w.b 4 "vlist_remove $w \[hlist_selection $w.h\]"
	mpack $w.b:bx $w.e:tx $w.h:tez
	entry_cmd $w.e "vlist_addname $w \[string trim \[entry_get $w.e\]\]"
}

proc vlist_watch {w sel} {
	set n [lindex $sel 0]
	hlist_insert $w.h $n "" [concat [lindex $sel 1] "1"]
	[hlist_subwidget $w.h text] tag configure $n -foreground red
}

proc vlist_remove {w sel} {
	set n [lindex $sel 0]
	hlist_clear $w.h $n/*
	hlist_delete $w.h $n
}

proc vlist_update {} {
	foreach w [window_list DeetVList] {
	  if {! [winfo exists $w]} return
	  foreach t [hlist_names $w.h] {
		set sel [hlist_selection $w.h $t]
		if {[lindex $sel 2] != ""} {
			set n [shpat_%% [hlist_text $w.h $t] { =.*}]
			hlist_clear $w.h $t/*
			hlist_delete $w.h $t
			if {[catch {set s [deet_sym -n $n]}]} continue
			vlist_add $w.h $t $n [lindex $s 1] [lindex $s 2] 1
			vlist_watch $w [list $t [hlist_selection $w.h $t]]
		}
	  }
	}
}

proc vlist_addname {w val} {
	set s [deet_sym -n $val]
	set id [lindex [split [[hlist_subwidget $w.h text] index end] .] 0]
	vlist_add $w.h $id $val [lindex $s 1] [lindex $s 2] 1
}

proc vlist_add {w id name tp val {expandSub 0}} {
	set gt [deet_type $tp]
	switch -- [lindex $gt 0] {
		pointer	{
			set pv [deet_getval $tp $val] 
			hlist_insert $w $id "$name = $pv" [list $tp $val]
			if { $pv != "0x0" && [lindex $gt 1] != "char" } {
				set nid $id/1
				if {[lindex [deet_type [lindex $gt 1]] 0] \
					== "struct"} {set nid $id}
				set cmd [list vlist_add $w $nid *$name \
					[lindex $gt 1] $pv 1]
				if {$expandSub} {eval $cmd} else {
					hlist_action $w $id $cmd
				}
				hlist_seticon $w $id plus
			}
		}
		union - struct {
			hlist_insert $w $id $name [list $tp $val]
			if {! $expandSub} {
				hlist_action $w $id [list vlist_add $w $id \
					$name $tp $val 1]
				hlist_seticon $w $id plus
				return
			}
			set v [deet_getval $tp $val]
			set s [lrange $gt 2 end]
			set i 0
			foreach s [lrange $gt 2 end] {
				if {[string match {\**} $name]} {
				  set elm "[shpat_# $name {\*}]->[lindex $s 1]"
				} else {
				  set elm "$name.[lindex $s 1]"
				}
				vlist_add $w $id/$i $elm [lindex $s 0] \
					[lindex $v $i]
				incr i
			}
		}
		array {
			hlist_insert $w $id $name [list $tp $val]
			if {! $expandSub} {
				hlist_action $w $id [list vlist_add $w $id \
					$name $tp $val 1]
				hlist_seticon $w $id plus
				return
			}
			set i 0
			foreach elm [deet_getval $tp $val] {
				vlist_add $w $id/$i "${name}\[$i\]" \
					[lindex $gt 1] $elm
				incr i
			}
		}
		default {
			hlist_insert $w $id "$name = [deet_getval $tp $val]" \
				[list $tp $val]
		}
	}
}

proc xprint_create {} {
	set w .xprint
	if {! [winfo exists $w]} {
		toplevel .xprint -class DeetVList
		window_title $w "Variables"
		vlist_setup $w
	}
	return $w
}

proc xprint {} {
	if {[catch {set val [selection get]}]} {
		return
	}
	set val [string trim $val]
	set w [xprint_create]
	vlist_addname $w $val
}

proc xdump {} {
	if {[catch {set frame [deet_frame]}]} return
	set w [window_create "" DeetVList]
	window_title $w "Dump: [lindex $frame 1]"
	vlist_setup $w
	set vals [concat [deet_sym -p] [deet_sym -l]]
	for {set i 0} {$vals != ""} {incr i} {
		vlist_add $w.h $i [shpat_## [lindex $vals 0] .*:] \
			[lindex $vals 1] [lindex $vals 2]
		set vals [lrange $vals 3 end]
	}
}

proc findfunc {name} {
	set n [lindex [deet_sym -n $name] 0]
	return [shpat_% $n :.*]
}

#####################################################################
# The below are unused right now, but will be used for interactive
# version
#####################################################################

proc dump {{arg -vars}} {
	if {[catch {set frame [deet_frame]}]} return
	set result ""
	if {$arg == "-args"} {
		set vals [deet_sym -p]
	} else {
		set vals [deet_sym -l]
	}
	while {$vals != ""} {
		append result "[lindex $vals 0]=[deet_getval [lindex $vals 1] [lindex $vals 2]]\n"
		set vals [lrange $vals 3 end]
	}
	return $result
}


proc var {name {tp ""}} {
	
}

proc ptype {ad} {
	set t [deet_type $ad]
	if {[lindex $t 0] == "pointer"} {
		set tp [ptype [lindex $t 1]]
		if {! [string match *\\* $tp]} {
			return "$tp *"
		} else {
			return "$tp*"
		}
	} elseif {[lindex $t 0] == "struct"} {
		return "struct [lindex $t 1]" 
	} else {
		return [lindex $t 0]
	}
}



proc pval {tp val} {
	set t [deet_type $tp]
	if {([lindex $t 0] == "struct") || ([lindex $t 0] == "union")} {
		puts "{"
		structVals arr $tp $val [lindex $t 2]
		foreach i [array names arr] {
			set a [arr($i)]
			puts "	[lindex $a 0] = [deet_getval [lindex $a 1] [lindex $a 2]]"
		}
		puts "}"
	} elseif {[lindex $t 0] == array} {
	 	print -nonewline "{ "
		foreach i [deet_getval $tp $val] {
			if {$arr != ""} {puts -nonewline ", "}
			puts -nonewline [deet_getval [lindex $t 1] $i]
			set arr 1
		}
		puts "}"
	} else {
		puts "([ptype $tp]) [deet_getval $tp $val]"
	}
}

#array is a nameref
proc structVals {array tp val {elements {}}} {
	set i 0

	if {$elements == ""} {
		set t [deet_type $tp]
		if {[lindex $t 0] == struct} {
			return 1
		}
		set elements [lindex $t 2]
	}

	set array [deet_getval $tp $val]
	set -- $elements
	while {[lindex $elements 0] && [lindex $elements 1]} {
		set array($i) [list [lindex $elements 1] [lindex $elements 0] $array($i)]
		incr i
		set elements [lrange $elements 2 end]
	}
	return 0
}
